@@ -26,6 +26,8 @@ SMTP queues will retry messages after certain periods of time.
 
 """
 
+from __future__ import absolute_import
+
 import time
 from itertools import imap, izip, repeat, chain
 
@@ -35,6 +37,7 @@ from gevent.event import Event
 from gevent.coros import Semaphore
 from gevent.pool import Pool
 
+from slimta.logging import log_exception
 from slimta.core import SlimtaError
 from slimta.relay import PermanentRelayError, TransientRelayError
 from slimta.smtp.reply import Reply
@@ -58,14 +61,14 @@ class QueueStorage(object):
         pass
 
     def write(self, envelope, timestamp):
-        """Writes the given envelope to storage, along with the timestamp of its
-        next delivery attempt. The number of delivery attempts asociated with
-        the message should start at zero.
+        """Writes the given envelope to storage, along with the timestamp of
+        its next delivery attempt. The number of delivery attempts asociated
+        with the message should start at zero.
 
         :param envelope: |Envelope| object to write.
         :param timestamp: Timestamp of message's next delivery attempt.
         :returns: Unique identifier string for the message in queue.
-        :raises: :class:`QueueError`.
+        :raises: :class:`QueueError`
 
         """
         raise NotImplementedError()
@@ -75,7 +78,7 @@ class QueueStorage(object):
 
         :param id: The unique identifier string for the message.
         :param timestamp: The new delivery attempt timestamp.
-        :raises: :class:`QueueError`.
+        :raises: :class:`QueueError`
 
         """
         raise NotImplementedError()
@@ -86,7 +89,7 @@ class QueueStorage(object):
 
         :param id: The unique identifier string for the message.
         :returns: The new number of message delivery attempts.
-        :raises: :class:`QueueError`.
+        :raises: :class:`QueueError`
 
         """
         raise NotImplementedError()
@@ -97,7 +100,7 @@ class QueueStorage(object):
 
         :returns: Iterable that yields tuples of the form ``(timestamp, id)``
                   for each message in storage.
-        :raises: :class:`QueueError`.
+        :raises: :class:`QueueError`
 
         """
         raise NotImplementedError()
@@ -106,7 +109,7 @@ class QueueStorage(object):
         """Returns the |Envelope| object associated with the given unique
         identifier sring.
 
-        :param id: The unique identfier string of the requested |Envelope|.
+        :param id: The unique identifier string of the requested |Envelope|.
         :returns: Tuple with the |Envelope| object and number of attempts.
         :raises: :class:`KeyError`, :class:`QueueError`
 
@@ -118,17 +121,41 @@ class QueueStorage(object):
         string from queue storage. This is typically used when the message has
         been successfully delivered or delivery has permanently failed.
 
-        :param id: The unique identfier string of the |Envelope| to remove.
-        :raises: :class:`QueueError`.
+        :param id: The unique identifier string of the |Envelope| to remove.
+        :raises: :class:`QueueError`
+
+        """
+        raise NotImplementedError()
+
+    def notify(self, id, timestamp):
+        """If messages are not received in the same process as queue delivery,
+        the storage mechanism needs a way to notify the queue delivery process
+        that a new message has been stored and is ready to deliver.
+
+        :param id: The unique identifier string of the |Envelope| that has
+                   already been stored and needs its first delivery attempt.
+        :param timestamp: Timestamp ofthe new message's next delivery attempt.
+
+        """
+        raise NotImplementedError()
+
+    def wait(self):
+        """If messages are not being delivered from the same process in which
+        they were received, the storage mechanism needs a way to wait until it
+        is notified that a new message has been stored.
+
+        :returns: A tuple of the timestamp and unique identifier string of a new
+                  message in storage. If ``None`` is returned indicating a
+                  timeout, this method is simply called again.
 
         """
         raise NotImplementedError()
 
 
 class Queue(Greenlet):
-    """Manages the queue of |Envelope| objects waiting for delivery. This is not
-    a standard FIFO queue, a message's place in the queue depends entirely on
-    the timestamp of its next delivery attempt.
+    """Manages the queue of |Envelope| objects waiting for delivery. This is
+    not a standard FIFO queue, a message's place in the queue depends entirely
+    on the timestamp of its next delivery attempt.
 
     :param store: Object implementing :class:`QueueStorage`.
     :param relay: |Relay| object used to attempt message deliveries.
@@ -139,17 +166,17 @@ class Queue(Greenlet):
                     returns ``None`` and messages are never retried.
     :param bounce_factory: Function that produces a |Bounce| object given the
                            same parameters as the |Bounce| constructor. If the
-                           function returns ``None``, no bounce is delivered. By
-                           default, a new |Bounce| is created in every case.
-    :param store_pool: Number of simultaneous operations performable against the
-                       ``store`` object. Default is unlimited.
-    :param relay_pool: Number of simultaneous operations performable against the
-                       ``relay`` object. Default is unlimited.
+                           function returns ``None``, no bounce is delivered.
+                           By default, a new |Bounce| is created in every case.
+    :param store_pool: Number of simultaneous operations performable against
+                       the ``store`` object. Default is unlimited.
+    :param relay_pool: Number of simultaneous operations performable against
+                       the ``relay`` object. Default is unlimited.
 
     """
 
     def __init__(self, store, relay, backoff=None, bounce_factory=None,
-                       store_pool=None, relay_pool=None):
+                 store_pool=None, relay_pool=None):
         super(Queue, self).__init__()
         self.store = store
         self.relay = relay
@@ -180,6 +207,7 @@ class Queue(Greenlet):
 
     def _run_policies(self, envelope):
         results = [envelope]
+
         def recurse(current, i):
             try:
                 policy = self.queue_policies[i]
@@ -227,7 +255,7 @@ class Queue(Greenlet):
 
     def _add_queued(self, entry):
         for i, info in enumerate(self.queued):
-            if info[0] > entry[0]: # [0] is the timestamp.
+            if info[0] > entry[0]:  # [0] is the timestamp.
                 self.queued.insert(i, entry)
                 break
         else:
@@ -235,15 +263,27 @@ class Queue(Greenlet):
         self.wake.set()
 
     def enqueue(self, envelope):
+        """Drops a new message in the queue for delivery. The first delivery
+        attempt is made immediately (depending on relay pool availability).
+        This method is not typically called directly, |Edge| objects use it
+        when they receive new messages.
+
+        :param envelope: |Envelope| object to enqueue.
+        :returns: Zipped list of envelopes and their respective queue IDs (or
+                  thrown :exc:`QueueError` objects).
+
+        """
         now = time.time()
         envelopes = self._run_policies(envelope)
-        ids = self._pool_imap('store', self.store.write, envelopes, repeat(now))
+        ids = self._pool_imap('store', self.store.write, envelopes,
+                              repeat(now))
         results = zip(envelopes, ids)
         for env, id in results:
             if not isinstance(id, BaseException):
-                self._pool_spawn('relay', self._attempt, id, env, 0)
+                self._pool_spawn('relay', self._notify_or_attempt,
+                                 id, now, env, 0)
             elif not isinstance(id, QueueError):
-                raise id # Re-raise exceptions that are not QueueError.
+                raise id  # Re-raise exceptions that are not QueueError.
         return results
 
     def _load_all(self):
@@ -257,7 +297,7 @@ class Queue(Greenlet):
 
     def _perm_fail(self, id, envelope, reply):
         self._pool_spawn('store', self.store.remove, id)
-        if envelope.sender: # Can't bounce to null-sender.
+        if envelope.sender:  # Can't bounce to null-sender.
             self._pool_spawn('bounce', self._bounce, envelope, reply)
 
     def _retry_later(self, id, envelope, reply):
@@ -271,13 +311,24 @@ class Queue(Greenlet):
             self.store.set_timestamp(id, when)
             self._add_queued((when, id))
 
+    def _notify_or_attempt(self, id, timestamp, envelope, attempts):
+        try:
+            self.store.notify(id, timestamp)
+        except NotImplementedError:
+            self._attempt(id, envelope, attempts)
+
     def _attempt(self, id, envelope, attempts):
         try:
             self.relay._attempt(envelope, attempts)
-        except TransientRelayError, e:
+        except TransientRelayError as e:
             self._pool_spawn('store', self._retry_later, id, envelope, e.reply)
-        except PermanentRelayError, e:
+        except PermanentRelayError as e:
             self._perm_fail(id, envelope, e.reply)
+        except Exception as e:
+            log_exception(__name__)
+            reply = Reply('450', '4.0.0 Unhandled delivery error: '+str(e))
+            self._pool_spawn('store', self._retry_later, id, envelope, reply)
+            raise
         else:
             self._pool_spawn('store', self.store.remove, id)
 
@@ -297,6 +348,15 @@ class Queue(Greenlet):
         if last_i > 0:
             self.queued = self.queued[last_i:]
 
+    def _wait_store(self):
+        while True:
+            try:
+                entry = self.store.wait()
+            except NotImplementedError:
+                return
+            if entry:
+                self._add_queued(entry)
+
     def _wait_ready(self, now):
         try:
             first = self.queued[0]
@@ -328,6 +388,7 @@ class Queue(Greenlet):
 
     def _run(self):
         self._pool_spawn('store', self._load_all)
+        self._pool_spawn('store', self._wait_store)
         while True:
             self.queued_lock.acquire()
             try:
