@@ -6,6 +6,7 @@ from functools import wraps
 from mox import MoxTestBase, IsA
 import gevent
 from gevent.pool import Pool
+from gevent.event import AsyncResult
 
 from slimta.queue import Queue, QueueStorage
 from slimta.smtp.reply import Reply
@@ -43,6 +44,8 @@ class TestQueue(MoxTestBase):
         self.assertRaises(NotImplementedError, qs.load)
         self.assertRaises(NotImplementedError, qs.get, '1234')
         self.assertRaises(NotImplementedError, qs.remove, '1234')
+        self.assertRaises(NotImplementedError, qs.notify, '1234', 0.0)
+        self.assertRaises(NotImplementedError, qs.wait)
 
     def test_policies(self):
         p1 = self.mox.CreateMock(QueuePolicy)
@@ -80,8 +83,17 @@ class TestQueue(MoxTestBase):
         self.assertEqual([], queue.queued)
         self.assertFalse(queue.wake.isSet())
 
+    def test_enqueue_notify(self):
+        self.store.write(self.env, IsA(float)).AndReturn('1234')
+        self.store.notify('1234', IsA(float))
+        self.mox.ReplayAll()
+        queue = Queue(self.store, self.relay, relay_pool=5)
+        self.assertEqual([(self.env, '1234')], queue.enqueue(self.env))
+        queue.relay_pool.join()
+
     def test_enqueue_wait(self):
         self.store.write(self.env, IsA(float)).AndReturn('1234')
+        self.store.notify('1234', IsA(float)).AndRaise(NotImplementedError)
         self.relay._attempt(self.env, 0)
         self.store.remove('1234')
         self.mox.ReplayAll()
@@ -104,6 +116,9 @@ class TestQueue(MoxTestBase):
         self.store.write(env1, IsA(float)).AndReturn('1234')
         self.store.write(env2, IsA(float)).AndReturn('5678')
         self.store.write(env3, IsA(float)).AndReturn('90AB')
+        self.store.notify('1234', IsA(float)).AndRaise(NotImplementedError)
+        self.store.notify('5678', IsA(float)).AndRaise(NotImplementedError)
+        self.store.notify('90AB', IsA(float)).AndRaise(NotImplementedError)
         self.relay._attempt(env1, 0).InAnyOrder('relay')
         self.relay._attempt(env2, 0).InAnyOrder('relay')
         self.relay._attempt(env3, 0).InAnyOrder('relay')
@@ -127,6 +142,7 @@ class TestQueue(MoxTestBase):
 
     def test_enqueue_wait_transientfail(self):
         self.store.write(self.env, IsA(float)).AndReturn('1234')
+        self.store.notify('1234', IsA(float)).AndRaise(NotImplementedError)
         self.relay._attempt(self.env, 0).AndRaise(TransientRelayError('transient', Reply('450', 'transient')))
         self.store.increment_attempts('1234')
         self.store.set_timestamp('1234', IsA(float))
@@ -139,6 +155,7 @@ class TestQueue(MoxTestBase):
 
     def test_enqueue_wait_transientfail_noretry(self):
         self.store.write(self.env, IsA(float)).AndReturn('1234')
+        self.store.notify('1234', IsA(float)).AndRaise(NotImplementedError)
         self.relay._attempt(self.env, 0).AndRaise(TransientRelayError('transient', Reply('450', 'transient')))
         self.store.increment_attempts('1234')
         self.store.remove('1234')
@@ -151,6 +168,7 @@ class TestQueue(MoxTestBase):
 
     def test_enqueue_wait_permanentfail(self):
         self.store.write(self.env, IsA(float)).AndReturn('1234')
+        self.store.notify('1234', IsA(float)).AndRaise(NotImplementedError)
         self.relay._attempt(self.env, 0).AndRaise(PermanentRelayError('permanent', Reply('550', 'permanent')))
         self.store.remove('1234')
         self.mox.ReplayAll()
@@ -163,6 +181,7 @@ class TestQueue(MoxTestBase):
     @_redirect_stderr
     def test_enqueue_wait_unhandledfail(self):
         self.store.write(self.env, IsA(float)).AndReturn('1234')
+        self.store.notify('1234', IsA(float)).AndRaise(NotImplementedError)
         self.relay._attempt(self.env, 0).AndRaise(Exception('unhandled error'))
         self.store.increment_attempts('1234')
         self.store.set_timestamp('1234', IsA(float))
@@ -190,6 +209,18 @@ class TestQueue(MoxTestBase):
         queue._check_ready(10)
         queue.store_pool.join()
 
+    def test_wait_store(self):
+        queue = Queue(self.store, self.relay, relay_pool=5)
+        queue.wake = self.mox.CreateMock(AsyncResult)
+        self.store.wait().AndReturn('1234')
+        queue.wake.set()
+        self.store.wait().AndReturn(None)
+        self.store.wait().AndReturn('5678')
+        queue.wake.set()
+        self.store.wait().AndRaise(NotImplementedError)
+        self.mox.ReplayAll()
+        queue._wait_store()
+
     def test_wait_ready_nonequeued(self):
         queue = Queue(self.store, self.relay)
         def wait_func():
